from functools import wraps
import inspect
from typing import Callable, Awaitable, NewType
from starlette.requests import Request
from starlette.responses import Response
import asyncio
from pypox._types import (
    QueryStr,
    QueryInt,
    QueryFloat,
    QueryBool,
    PathStr,
    PathInt,
    PathFloat,
    PathBool,
    HeaderStr,
    HeaderInt,
    HeaderFloat,
    HeaderBool,
    CookieStr,
    CookieInt,
    CookieFloat,
    CookieBool,
    BodyDict,
)


def processor(func: Callable) -> Callable:

    @wraps(func)
    async def wrapper(request: Request) -> Response:
        params = await load_parameters(func, request)
        if inspect.iscoroutinefunction(func):
            return await func(**params)
        else:
            return func(**params)

    return wrapper


async def load_parameters(func: Callable[[], Awaitable], request: Request) -> dict:

    params = {}

    for name, annotation in inspect.signature(func).parameters.items():
        if annotation.annotation == Request:
            params[name] = request
        if annotation.annotation in [QueryStr, QueryInt, QueryFloat, QueryBool]:
            if name.replace("_", "-") in request.query_params:
                params[name] = validate_query(
                    request.query_params[name.replace("_", "-")], annotation.annotation
                )
            else:
                params[name] = validate_query(
                    request.query_params[name.replace("-", "_")], annotation.annotation
                )
        if annotation.annotation in [PathStr, PathInt, PathFloat, PathBool]:
            if name.replace("_", "-") in request.path_params:
                params[name] = validate_path(
                    request.path_params[name.replace("_", "-")], annotation.annotation
                )
            else:
                params[name] = validate_path(
                    request.path_params[name.replace("-", "_")], annotation.annotation
                )
        if annotation.annotation in [HeaderStr, HeaderInt, HeaderFloat, HeaderBool]:
            params[name] = validate_header(
                request.headers[name.replace("_", "-")], annotation.annotation
            )
        if annotation.annotation in [CookieStr, CookieInt, CookieFloat, CookieBool]:
            params[name] = validate_cookie(
                request.cookies[name.replace("_", "-")], annotation.annotation
            )
        if annotation.annotation in [BodyDict]:
            params[name] = await request.json()
        await asyncio.sleep(0)
    return params


def validate_query(value, _type):
    if _type == QueryStr:
        return str(value)
    if _type == QueryInt:
        return int(value)
    if _type == QueryFloat:
        return float(value)
    if _type == QueryBool:
        return bool(value)


def validate_path(value, _type):
    if _type == PathStr:
        return str(value)
    if _type == PathInt:
        return int(value)
    if _type == PathFloat:
        return float(value)
    if _type == PathBool:
        return bool(value)


def validate_header(value, _type):
    if _type == HeaderStr:
        return str(value)
    if _type == HeaderInt:
        return int(value)
    if _type == HeaderFloat:
        return float(value)
    if _type == HeaderBool:
        return bool(value)


def validate_cookie(value, _type):
    if _type == CookieStr:
        return str(value)
    if _type == CookieInt:
        return int(value)
    if _type == CookieFloat:
        return float(value)
    if _type == CookieBool:
        return bool(value)



from contextlib import AbstractAsyncContextManager
from importlib.machinery import ModuleSpec
from typing import Any, Callable, Generator, Sequence
from starlette.middleware import Middleware
from starlette.routing import BaseRoute, Router, Route, WebSocketRoute
from starlette.types import ASGIApp
from typing import MutableMapping, Awaitable, Mapping
from types import ModuleType
import importlib.util
import os


class BaseRouter(Router):

    def __init__(
        self,
        directory: str,
        entry_point: str,
        file: dict[str, str],
        _type: str = "http",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None
    ) -> None:

        self._router_type = _type
        self._callable = entry_point
        self._directory = directory
        self._file = file
        super().__init__(
            self.generate_routes(),
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )

    @property
    def directory(self) -> str:
        return self._directory

    @property
    def callable(self) -> str:
        return self._callable

    @property
    def router_type(self) -> str:
        return self._router_type

    def generate_routes(self) -> list[BaseRoute]:
        router: list[BaseRoute] = []

        for root, file in self.walk():
            module_name: str = file.split(".")[0]
            module_path: str = os.path.join(root, file)

            module: ModuleType = self.load_module(module_name, module_path)

            route_path: str = self.create_route_path(self.directory, root)
            router.append(self.create_route(route_path, getattr(module, self.callable), methods=[self._file[file]]))

        return router

    def walk(self) -> Generator[tuple[str, str], Any, None]:
        """Recursively walks through the directory and yields tuples of root and file names.

        Yields:
            Generator[tuple[str, str], Any, None]: A generator that yields tuples of root and file names.
        """
        for root, _, files in os.walk(self.directory):
            for file in files:
                if file in self._file:
                    yield root, file

    def load_module(self, module_name: str, module_path: str) -> ModuleType:
        """
        Load a module from a file location.

        Args:
            module_name (str): The name of the module.
            module_path (str): The path to the module file.

        Raises:
            ModuleNotFoundError: If the module is not found.
            AttributeError: If the callable is not found in the module.

        Returns:
            ModuleType: The loaded module.
        """
        spec: ModuleSpec | None = importlib.util.spec_from_file_location(
            module_name, module_path
        )
        if not spec:
            raise ModuleNotFoundError("Module not found")
        module: ModuleType = importlib.util.module_from_spec(spec)
        if not spec.loader:
            raise ModuleNotFoundError("Module not found")
        spec.loader.exec_module(module)
        if not hasattr(module, self.callable):
            raise AttributeError("Callable not found in module")
        return module

    def create_route_path(self, directory: str, root: str) -> str:
        """
        Create a route path by replacing the directory in the root path and performing some string replacements.

        Args:
            directory (str): The directory to be replaced in the root path.
            root (str): The root path.

        Returns:
            str: The modified route path.
        """
        return (
            root.replace(directory, "")
            .replace("\\", "/")
            .replace("[", "{")
            .replace("]", "}")
            + "/"
        )

    def create_route(
        self, route_path: str, func: Callable, methods: list[str] = ["GET"]
    ) -> Route | WebSocketRoute:
        """
        Create a route with the given path and endpoint.

        Args:
            route_path (str): The path of the route.
            endpoint (Callable): The endpoint of the route.

        Returns:
            BaseRoute: The created route.
        """

        if self.router_type == "http":
            return Route(route_path, func, methods=methods)

        if self.router_type == "websocket":
            return WebSocketRoute(route_path, func, name=func.__name__)

        raise ValueError("Invalid router type")


class HTTPRouter(BaseRouter):

    def __init__(
        self,
        directory: str,
        entry_point: str = "endpoint",
        file: dict[str, str] = {},
        _type: str = "http",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None
    ) -> None:

        if not file:
            file = {
                "get.py": "GET",
                "post.py": "POST",
                "put.py": "PUT",
                "delete.py": "DELETE",
                "patch.py": "PATCH",
                "head.py": "HEAD",
                "options.py": "OPTIONS",
            }

        super().__init__(
            directory,
            entry_point,
            file,
            _type,
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )


class WebsocketRouter(BaseRouter):

    def __init__(
        self,
        directory: str,
        entry_point: str = "endpoint",
        file: dict[str, str] = {},
        _type: str = "websocket",
        redirect_slashes: bool = True,
        default: (
            Callable[
                [
                    MutableMapping[str, Any],
                    Callable[[], Awaitable[MutableMapping[str, Any]]],
                    Callable[[MutableMapping[str, Any]], Awaitable[None]],
                ],
                Awaitable[None],
            ]
            | None
        ) = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: (
            Callable[[Any], AbstractAsyncContextManager[None]]
            | Callable[[Any], AbstractAsyncContextManager[Mapping[str, Any]]]
            | None
        ) = None,
        *,
        middleware: Sequence[Middleware] | None = None
    ) -> None:

        if not file:
            file = {"websocket.py": "WEBSOCKET"}
        super().__init__(
            directory,
            entry_point,
            file,
            _type,
            redirect_slashes,
            default,
            on_startup,
            on_shutdown,
            lifespan,
            middleware=middleware,
        )



from typing import Literal


class HTTP_100_CONTINUE:
    code = 100


class HTTP_101_SWITCHING_PROTOCOLS:
    code = 101


class HTTP_102_PROCESSING:
    code = 102


class HTTP_103_EARLY_HINTS:
    code = 103


class HTTP_200_OK:
    code = 200


class HTTP_201_CREATED:
    code = 201


class HTTP_202_ACCEPTED:
    code = 202


class HTTP_203_NON_AUTHORITATIVE_INFORMATION:
    code = 203


class HTTP_204_NO_CONTENT:
    code = 204


class HTTP_205_RESET_CONTENT:
    code = 205


class HTTP_206_PARTIAL_CONTENT:
    code = 206


class HTTP_207_MULTI_STATUS:
    code = 207


class HTTP_208_ALREADY_REPORTED:
    code = 208


class HTTP_226_IM_USED:
    code = 226


class HTTP_300_MULTIPLE_CHOICES:
    code = 300


class HTTP_301_MOVED_PERMANENTLY:
    code = 301


class HTTP_302_FOUND:
    code = 302


class HTTP_303_SEE_OTHER:
    code = 303


class HTTP_304_NOT_MODIFIED:
    code = 304


class HTTP_305_USE_PROXY:
    code = 305


class HTTP_306_RESERVED:
    code = 306


class HTTP_307_TEMPORARY_REDIRECT:
    code = 307


class HTTP_308_PERMANENT_REDIRECT:
    code = 308


class HTTP_400_BAD_REQUEST:
    code = 400


class HTTP_401_UNAUTHORIZED:
    code = 401


class HTTP_402_PAYMENT_REQUIRED:
    code = 402


class HTTP_403_FORBIDDEN:
    code = 403


class HTTP_404_NOT_FOUND:
    code = 404


class HTTP_405_METHOD_NOT_ALLOWED:
    code = 405


class HTTP_406_NOT_ACCEPTABLE:
    code = 406


class HTTP_407_PROXY_AUTHENTICATION_REQUIRED:
    code = 407


class HTTP_408_REQUEST_TIMEOUT:
    code = 408


class HTTP_409_CONFLICT:
    code = 409


class HTTP_410_GONE:
    code = 410


class HTTP_411_LENGTH_REQUIRED:
    code = 411


class HTTP_412_PRECONDITION_FAILED:
    code = 412


class HTTP_413_REQUEST_ENTITY_TOO_LARGE:
    code = 413


class HTTP_414_REQUEST_URI_TOO_LONG:
    code = 414


class HTTP_415_UNSUPPORTED_MEDIA_TYPE:
    code = 415


class HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE:
    code = 416


class HTTP_417_EXPECTATION_FAILED:
    code = 417


class HTTP_418_IM_A_TEAPOT:
    code = 418


class HTTP_421_MISDIRECTED_REQUEST:
    code = 421


class HTTP_422_UNPROCESSABLE_ENTITY:
    code = 422


class HTTP_423_LOCKED:
    code = 423


class HTTP_424_FAILED_DEPENDENCY:
    code = 424


class HTTP_425_TOO_EARLY:
    code = 425


class HTTP_426_UPGRADE_REQUIRED:
    code = 426


class HTTP_428_PRECONDITION_REQUIRED:
    code = 428


class HTTP_429_TOO_MANY_REQUESTS:
    code = 429


class HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE:
    code = 431


class HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS:
    code = 451


class HTTP_500_INTERNAL_SERVER_ERROR:
    code = 500


class HTTP_501_NOT_IMPLEMENTED:
    code = 501


class HTTP_502_BAD_GATEWAY:
    code = 502


class HTTP_503_SERVICE_UNAVAILABLE:
    code = 503


class HTTP_504_GATEWAY_TIMEOUT:
    code = 504


class HTTP_505_HTTP_VERSION_NOT_SUPPORTED:
    code = 505


class HTTP_506_VARIANT_ALSO_NEGOTIATES:
    code = 506


class HTTP_507_INSUFFICIENT_STORAGE:
    code = 507


class HTTP_508_LOOP_DETECTED:
    code = 508


class HTTP_510_NOT_EXTENDED:
    code = 510


class HTTP_511_NETWORK_AUTHENTICATION_REQUIRED:
    code = 511


class WS_1000_NORMAL_CLOSURE:
    code = 1000


class WS_1001_GOING_AWAY:
    code = 1001


class WS_1002_PROTOCOL_ERROR:
    code = 1002


class WS_1003_UNSUPPORTED_DATA:
    code = 1003


class WS_1005_NO_STATUS_RCVD:
    code = 1005


class WS_1006_ABNORMAL_CLOSURE:
    code = 1006


class WS_1007_INVALID_FRAME_PAYLOAD_DATA:
    code = 1007


class WS_1008_POLICY_VIOLATION:
    code = 1008


class WS_1009_MESSAGE_TOO_BIG:
    code = 1009


class WS_1010_MANDATORY_EXT:
    code = 1010


class WS_1011_INTERNAL_ERROR:
    code = 1011


class WS_1012_SERVICE_RESTART:
    code = 1012


class WS_1013_TRY_AGAIN_LATER:
    code = 1013


class WS_1014_BAD_GATEWAY:
    code = 1014


class WS_1015_TLS_HANDSHAKE:
    code = 1015



type QueryStr = str
type QueryInt = int
type QueryFloat = float
type QueryBool = bool
type PathStr = str
type PathInt = int
type PathFloat = float
type PathBool = bool
type HeaderStr = str
type HeaderInt = int
type HeaderFloat = float
type HeaderBool = bool
type CookieStr = str
type CookieInt = int
type CookieFloat = float
type CookieBool = bool
type BodyDict = dict



import os
from typing import Any, Callable, Mapping, Sequence, TypeVar, ParamSpec
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.routing import Route, Router, Mount, BaseRoute
from starlette.types import ExceptionHandler, Lifespan, Receive, Scope, Send
from pypox.router import BaseRouter
from pypox.openapi.main import OpenAPI, Info, License


class Pypox(Starlette):
    """A web application framework based on Starlette.

    Args:
        Starlette (type): The base class for the Pypox application.

    Attributes:
        processor_func (list[BaseProcessor] | None): The list of processor functions.
        conventions (list[BaseConvention]): The list of conventions.
        debug (bool): Flag indicating whether debug mode is enabled.
        middleware (Sequence[Middleware] | None): The middleware stack.
        exception_handlers (Mapping[Any, ExceptionHandler] | None): The exception handlers.
        on_startup (Sequence[Callable[[], Any]] | None): The startup functions.
        on_shutdown (Sequence[Callable[[], Any]] | None): The shutdown functions.
        lifespan (Lifespan | None): The lifespan of the application.
    """

    def __init__(
        self,
        conventions: list[BaseRouter],
        debug: bool = False,
        middleware: Sequence[Middleware] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan | None = None,
        open_api_version: str = "3.0.3",
        info: Info = Info(title="Pypox", version="0.0.1", license=License(name="MIT")),
        license: License = License(name="MIT"),
    ) -> None:
        self._openapi_version = open_api_version
        self._info = info
        self._license = license
        routes: list[BaseRoute] = []
        for convention in conventions:
            routes.extend(convention.generate_routes())
        super().__init__(
            debug,
            routes,
            middleware,
            exception_handlers,
            on_startup,
            on_shutdown,
            lifespan,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["openapi"] = {
            "openapi": self._openapi_version,
            "info": self._info,
            "license": self._license,
        }
        return await super().__call__(scope, receive, send)
